@@ -5,9 +5,8 @@
 
 struct wireguard_peer;
 
-void timers_init_peer(struct wireguard_peer *peer);
-void timers_uninit_peer(struct wireguard_peer *peer);
-
+void timers_init(struct wireguard_peer *peer);
+void timers_stop(struct wireguard_peer *peer);
 void timers_data_sent(struct wireguard_peer *peer);
 void timers_data_received(struct wireguard_peer *peer);
 void timers_any_authenticated_packet_received(struct wireguard_peer *peer);
